@@ -108,7 +108,7 @@ srs_startElement (void           *ctx,
 				}
 				if ((last_priority < srs_priority) ||
 				    ((last_priority == srs_priority) && last_preempt))
-				    srs_call_shut_up ();
+				    srs_call_shut_up (srs_priority, srs_preempt);
 				last_priority = srs_priority;
 				last_preempt = srs_preempt;
 				srs_remove (srs_priority);						
@@ -497,7 +497,7 @@ static struct
 	{'-',	N_("dash")	},
     };
 
-
+/* To translators: This is the military phonetic alphabet */
 static struct
 {
     gchar ch;
