@@ -510,7 +510,7 @@ src_speech_say_ (gchar *text_,
     if  (attr)
 	text = src_xml_format ("TEXT", attr, text_);
     if (text)
-	rv = src_speech_send_chunk (text, priority);
+	rv = src_speech_send_chunk (text, priority, TRUE);
     g_free (text);
     g_free (attr);
 
@@ -1037,20 +1037,17 @@ src_speech_resume ()
 
 
 gboolean
-src_speech_send_chunk (gchar *chunk, SRCSpeechPriority priority)
+src_speech_send_chunk (gchar *chunk, SRCSpeechPriority priority, gboolean preempt)
 {
     gchar *output;
     gboolean rv = FALSE;
-    static SRCSpeechPriority last_priority = SRC_SPEECH_PRIORITY_IDLE;
 
     sru_assert (src_speech_initialized);
     sru_assert (chunk && chunk[0]);
-    if (last_priority >= priority)
-	src_speech_shutup ();
-    output = g_strconcat ("<SRSOUT>", chunk, "</SRSOUT>", NULL);
+    output = g_strdup_printf ("<SRSOUT priority=\"%d\" preempt=\"%s\">%s</SRSOUT>",
+			priority, preempt ? "true" : "false", chunk);
     rv = src_speech_send (output);
     g_free (output);
-    last_priority = priority;
 
     return rv;
 }
@@ -1150,8 +1147,9 @@ src_dictionary_load ()
 	
 	crt = crt->next;
     }
-  	        
-    srconf_free_slist (dictionary_list);
+ 
+    if (dictionary_list)
+	srconf_free_slist (dictionary_list);
     
     return TRUE;
 }
@@ -1513,6 +1511,9 @@ src_speech_process_string (const gchar   *text,
 	
     if (text[0] == ' ' && text[1] == '\0' && smode == SRC_SPEECH_SPELL_NORMAL)
 	return g_strdup (_("space"));
+    if (text[0] == '\n' && text[1] == '\0' && smode == SRC_SPEECH_SPELL_NORMAL)
+	return g_strdup (("enter")); /*FIXME: should be marked for translation */
+
     if (text[0] && !text[1])
     {
         gint index = src_speech_get_index_for_symbol (crt[0]);
