@@ -79,6 +79,8 @@ typedef enum _SRLEventType
     SRL_EVENT_FOCUS1, /* some events are reported as "focus:",
 			but all other events for those objects 
 			should not be reported. (eg ALT+CTRL+TAB) */
+    SRL_EVENT_FOCUS2, /* some text object are not focused, but user can 
+			interact with them */
     SRL_EVENT_LINK_SELECTED,
     SRL_EVENT_TEXT_CARET_MOVED,
     SRL_EVENT_TEXT_CHANGED_INSERT,
@@ -174,7 +176,8 @@ static gboolean			srl_idle_installed	= FALSE;
 static GQueue			*srl_event_queue	= NULL;
 static AccessibleEventListener 	*srl_event_listeners[G_N_ELEMENTS (srl_events_type_name)];
 static Accessible		*srl_last_focus		= NULL;
-static GPtrArray		*srl_watched_acc	= NULL;
+static Accessible		*srl_watched_acc	= NULL;
+static Accessible 		*srl_last_table 	= NULL;
 /*FIXME:next varaible should be in SRObject.c file */
 Accessible 			*srl_last_edit		= NULL;
 
@@ -183,6 +186,8 @@ extern gboolean sro_set_name (SRObject *obj, gchar *name);
 extern gboolean sro_get_from_accessible_event (Accessible *acc, gchar *reason, 
 						SRObject **obj);
 extern Accessible* sro_get_acc (SRObject *obj);
+extern gint sr_acc_get_link_index (Accessible *acc);
+
 #define sro_get_from_acc	sro_get_from_accessible_event
 
 #ifdef SRL_TIMING
@@ -677,6 +682,8 @@ srle_get_reason (SRLEvent *event)
 	return SRL_EVENT_NAME_WINDOW_RENAME;
     if (srle_has_type (event, SRL_EVENT_FOCUS1))
 	return SRL_EVENT_NAME_FOCUS;
+    if (srle_has_type (event, SRL_EVENT_FOCUS2))
+	return SRL_EVENT_NAME_FOCUS;
 
     srl_assert_not_reached ();
     return SRL_EVENT_NAME_UNKNOWN;
@@ -687,6 +694,8 @@ srl_is_window_event (SRLEvent *event)
 {
     srl_assert (event);
     
+    if (srle_acc_has_role (event, SPI_ROLE_TOOL_TIP))
+	return FALSE;
     return  event->type == SRL_EVENT_WINDOW_CREATE   ||
 	    event->type == SRL_EVENT_WINDOW_DESTROY  ||
 	    event->type == SRL_EVENT_WINDOW_MINIMIZE ||
@@ -705,6 +714,7 @@ srl_is_focus_event (SRLEvent *event)
     
     return  event->type == SRL_EVENT_FOCUS		||
 	    event->type == SRL_EVENT_FOCUS1		||
+	    event->type == SRL_EVENT_FOCUS2		||
 	    event->type == SRL_EVENT_LINK_SELECTED	||
 	    event->type == SRL_EVENT_TEXT_CARET_MOVED	||
 	    event->type == SRL_EVENT_TEXT_CHANGED_INSERT||
@@ -724,8 +734,7 @@ srl_is_tooltip_event (SRLEvent *event)
 {
     srl_assert (event);
     
-    return  event->type == SRL_EVENT_TOOLTIP_SHOW	||
-	    event->type == SRL_EVENT_TOOLTIP_HIDE;
+    return  event->type == SRL_EVENT_TOOLTIP_SHOW;
 }
 
 static gboolean
@@ -768,13 +777,70 @@ srl_notify_all_clients (SREvent *event,
     return TRUE;
 }
 
+static gboolean
+srl_set_last_table (Accessible *acc)
+{
+    if (srl_last_table)
+	Accessible_unref (srl_last_table);
+    srl_last_table = acc;
+    if (srl_last_table)
+	Accessible_ref (srl_last_table);
+    return TRUE;
+}
+
+static gboolean
+srl_table_same_line (Accessible *acc)
+{
+    static gint last_row = -1;
+    Accessible *parent;
+    gboolean rv = FALSE;
+
+    srl_assert (acc);
+
+    parent = Accessible_getParent (acc);
+    if (parent)
+    {
+	AccessibleTable *table= Accessible_getTable (parent);
+	if (table)
+	{
+	    gint row = AccessibleTable_getRowAtIndex (table, 
+				    Accessible_getIndexInParent (acc));
+	    if (row == last_row && parent == srl_last_table)
+	        rv = TRUE;
+	    last_row = row;
+	    AccessibleTable_unref (table);
+	}
+	Accessible_unref (parent);
+    }
+    return rv;
+}
+
+
+static gboolean
+srl_table_same_header (Accessible *acc)
+{
+    Accessible *parent;
+    gboolean rv = FALSE;
+
+    srl_assert (acc);
+
+    parent = Accessible_getParent (acc);
+    if (parent)
+    {
+	if (parent == srl_last_table)
+	    rv = TRUE;
+	Accessible_unref (parent);
+    }
+    return rv;
+}
+
 static gboolean
 srl_notify_clients_obj (SRLEvent *event,
 			SREventType type)
 {
     SREvent *ev;
     SRObject *obj;
-
+    
     srl_assert (event);
 
     ev = sre_new ();
@@ -784,6 +850,66 @@ srl_notify_clients_obj (SRLEvent *event,
 				    (gchar*)srle_get_reason (event), &obj))
 	obj = NULL;
 
+    if (obj)
+    {
+	SRObjectRoles role;
+	sro_get_role (obj, &role, SR_INDEX_CONTAINER);
+	
+	if ((role == SR_ROLE_TABLE_LINE || role == SR_ROLE_TABLE_COLUMNS_HEADER) &&
+	    srle_has_type (event, SRL_EVENT_FOCUS))
+	{
+	    SRObject *pobj = NULL;
+	    Accessible *pacc;
+ 
+	    pacc = Accessible_getParent ((Accessible*)srle_get_acc (event));
+
+	    if (pacc != srl_last_table)
+	    {
+		if (pacc)
+		    if (!sro_get_from_acc (pacc, SRL_EVENT_NAME_FOCUS, &pobj))
+			pobj = NULL;
+		if (pobj)
+		{
+		    SREvent *pev = NULL;
+		    pev = sre_new ();
+		    pev->type = type;
+		    pev->data = pobj;
+    		    pev->data_destructor = (SREventDataDestructor) sro_release_reference;
+		    srl_notify_all_clients (pev, 0);
+		    sre_release_reference (pev);
+		}
+	    }
+	    if (pacc)
+		Accessible_unref (pacc);
+	}
+    }
+
+    if (obj)
+    {
+	static SRObjectRoles last_role = SR_ROLE_UNKNOWN; /* to report all headers
+		in case of click from a table cell to a column header */
+	SRObjectRoles role;
+	sro_get_role (obj, &role, SR_INDEX_CONTAINER);
+	if (srle_has_type (event, SRL_EVENT_FOCUS))
+	    if ((role == SR_ROLE_TABLE_LINE && 
+		    srl_table_same_line ((Accessible*)srle_get_acc (event))) ||
+	        (role == SR_ROLE_TABLE_COLUMNS_HEADER && role == last_role &&
+	    	    srl_table_same_header ((Accessible*)srle_get_acc (event))))
+		{
+		    sro_release_reference (obj);
+		    if (!sro_get_from_acc ((Accessible*)srle_get_acc (event),
+				SRL_EVENT_NAME_SELECTION_CHANGED, &obj))
+			obj = NULL;
+		}
+	if (role == SR_ROLE_TABLE_LINE || role == SR_ROLE_TABLE_COLUMNS_HEADER)
+	{
+	    Accessible *parent = Accessible_getParent ((Accessible*)srle_get_acc (event));
+	    srl_set_last_table (parent);
+	    Accessible_unref (parent);
+	    last_role = role;
+	}
+    }
+
     if (obj)
     {
 	ev->type = type;
@@ -1052,9 +1178,17 @@ srle_change_type (SRLEvent *event)
     }
     else if (srle_has_type (event, SRL_EVENT_TEXT_CARET_MOVED))
     {
-	if (!srle_is_for_focused_acc (event) &&
+	static gint last_link_index = -1;
+	gint new_link_index = sr_acc_get_link_index ((Accessible*)srle_get_acc (event));
+/*	if (!srle_is_for_focused_acc (event) &&
 	    srl_acc_has_toolkit (srle_get_acc (event), SRL_TOOLKIT_MOZILLA))
 	    event->type = SRL_EVENT_FOCUS;
+*/
+	if (!srle_is_for_focused_acc (event))
+	    event->type = SRL_EVENT_FOCUS2;
+	if (last_link_index != new_link_index)
+	    event->type = SRL_EVENT_FOCUS;
+	last_link_index = new_link_index;
     }
     else if (srle_has_type (event, SRL_EVENT_NAME_CHANGED))
     {
@@ -1077,7 +1211,7 @@ srle_change_type (SRLEvent *event)
 	if (srl_acc_has_toolkit (srle_get_acc (event), SRL_TOOLKIT_MOZILLA))
 	    event->type = SRL_EVENT_FOCUS;
     }
-
+    
     return TRUE;
 }
 
@@ -1215,7 +1349,7 @@ srl_text_event_is_reported (SRLEvent *event,
 	rv = TRUE;
 
     if (srle_acc_has_role (event, SPI_ROLE_SPIN_BUTTON))
-        rv = srl_info_is_value_changed (info, info2);
+        rv = !srl_info_is_value_changed (info, info2);
 
     /* when a text-changed event for child of combos is emitted,
        if there is a selection,then this change take place 
@@ -1298,17 +1432,11 @@ srl_selection_event_is_reported (SRLEvent *event)
 }
 
 static gboolean
-srle_acc_is_watched (SRLEvent *event)
+srle_is_for_watched_acc (SRLEvent *event)
 {
-    gint i;
-
     srl_assert (event);
     
-    for (i = 0; i < srl_watched_acc->len; i++)
-	if (event->acc_ev->source == g_ptr_array_index (srl_watched_acc, i))
-	    return TRUE;
-
-    return FALSE;
+    return srl_watched_acc == event->acc_ev->source;
 }
 
 
@@ -1316,9 +1444,11 @@ static gboolean
 srl_event_is_reported (SRLEvent *event)
 {
     gboolean rv = FALSE;
-    static SRLLastInfo last_info;
+    static SRLLastInfo last_info_focus, last_info_watched;
+    SRLLastInfo *last_info = NULL;
 
     srl_assert (event);
+    last_info = srle_is_for_focused_acc (event) ? &last_info_focus : &last_info_watched;
 
     if (srl_is_window_event (event) 	||
 	srl_is_mouse_event (event)	||
@@ -1328,7 +1458,10 @@ srl_event_is_reported (SRLEvent *event)
 	rv = srle_acc_has_state (event, SPI_STATE_FOCUSED);
     else if (srle_has_type (event, SRL_EVENT_FOCUS1))
 	rv = TRUE;
-    else if (srle_is_for_focused_acc (event))
+    else if (srle_has_type (event, SRL_EVENT_FOCUS2))
+	rv = TRUE;
+    else if (srle_is_for_focused_acc (event) ||
+	     srle_is_for_watched_acc (event))
     {
 	if (srle_has_type (event, SRL_EVENT_STATE_EXPANDED)) 
 	    rv = srle_acc_has_role (event, SPI_ROLE_TABLE_CELL);
@@ -1346,7 +1479,7 @@ srl_event_is_reported (SRLEvent *event)
 		 srle_has_type (event, SRL_EVENT_TEXT_CHANGED_DELETE) ||
 		 srle_has_type (event, SRL_EVENT_TEXT_CARET_MOVED)    ||
 		 srle_has_type (event, SRL_EVENT_TEXT_SELECTION_CHANGED))
-	    rv = srl_text_event_is_reported (event, last_info);
+	    rv = srl_text_event_is_reported (event, *last_info);
 	else if (srle_has_type (event, SRL_EVENT_SELECTION_CHANGED))
 	    rv = srl_selection_event_is_reported (event);
 	else if (srle_has_type (event, SRL_EVENT_NAME_CHANGED))
@@ -1359,27 +1492,23 @@ srl_event_is_reported (SRLEvent *event)
     if (srle_has_type (event, SRL_EVENT_WINDOW_DEACTIVATE))
 	rv = FALSE;
 
-    if (rv && srle_is_for_focused_acc (event))
+    if (rv && (srle_is_for_focused_acc (event) ||
+	      srle_is_for_watched_acc (event)))
     {
-	srl_set_info (event->acc_ev->source, &last_info);
+	srl_assert (last_info);
+	srl_set_info (event->acc_ev->source, last_info);
 
-	/* for text-changed:delete event information from object contains
-	   values from before event was fired, not the present */
-	if (srle_has_type (event, SRL_EVENT_TEXT_CHANGED_DELETE))
-	{
-	    last_info.text_caret_offset     = event->acc_ev->detail1;
-	    last_info.text_selections_count = 0;
-	    last_info.text_char_count      -= event->acc_ev->detail2;
-	    last_info.text_crt_selection_length = 0;
-	}
+	if (srle_is_for_focused_acc (event))
+	    if (last_info != &last_info_focus)
+		last_info_focus = *last_info;
+	if (srle_is_for_watched_acc (event))
+	    if (last_info != &last_info_watched)
+		last_info_watched = *last_info;
     }
     
     if (srle_acc_has_role (event, SPI_ROLE_EDITBAR))
 	rv = TRUE;
 
-    if (srle_acc_is_watched (event))
-	rv = TRUE;
-
     return rv;
 }
 
@@ -1404,17 +1533,17 @@ srl_process_event (SRLEvent *event)
     srl_timing_process_start ((gpointer) event->acc_ev);
     srle_set_acc (event);
     srle_change_type (event);
-    process = srl_event_is_reported (event);
 
-    if (srle_has_type (event, SRL_EVENT_FOCUS))
-	srl_set_last_focus ((Accessible*)srle_get_acc (event));
+    if (srle_has_type (event, SRL_EVENT_FOCUS) ||
+	srle_has_type (event, SRL_EVENT_FOCUS2))
+	    srl_set_last_focus ((Accessible*)srle_get_acc (event));
     else if (srle_has_type (event, SRL_EVENT_FOCUS1))
 	srl_set_last_focus (NULL);
 
 /* FIXME: 
     if (srle_has_type (event, SRL_EVENT_WINDOW_DEACTIVATE))
 	report no focus if no other event */
-
+    process = srl_event_is_reported (event);
     if (process)
 	srl_report_event (event);
 
@@ -1487,7 +1616,8 @@ srl_init ()
     srl_last_edit	= NULL;
     srl_idle_need	= FALSE;
     srl_idle_installed	= FALSE;
-    srl_watched_acc	= g_ptr_array_sized_new (10);
+    srl_watched_acc	= NULL;
+    srl_last_table	= NULL;
     
     srl_log_init ();
     srl_timing_init ();
@@ -1544,9 +1674,10 @@ srl_terminate ()
 	Accessible_unref (srl_last_focus);
     if (srl_last_edit)
         Accessible_unref (srl_last_edit);
-    for (i = 0; i < srl_watched_acc->len; i++)
-	Accessible_unref (g_ptr_array_index (srl_watched_acc, i));
-    g_ptr_array_free (srl_watched_acc, TRUE);
+    if (srl_watched_acc)
+	Accessible_unref (srl_watched_acc);
+    if (srl_last_table)
+	Accessible_unref (srl_last_table);
     srl_log_terminate ();
     srl_timing_terminate ();
     srl_set_uninitialized ();
@@ -1662,7 +1793,9 @@ srl_set_watch_for_object (SRObject *obj)
     srl_unwatch_all_objects ();
     acc = sro_get_acc (obj);
     Accessible_ref (acc);
-    g_ptr_array_add (srl_watched_acc, acc);
+    if (srl_watched_acc)
+	Accessible_unref (srl_watched_acc);
+    srl_watched_acc = acc;
 
     return TRUE;
 }
@@ -1670,12 +1803,7 @@ srl_set_watch_for_object (SRObject *obj)
 void
 srl_unwatch_all_objects ()
 {
-    gint i;
-    
-    for (i = 0; i < srl_watched_acc->len; i++)
-    {
-	Accessible *acc = g_ptr_array_index (srl_watched_acc, i);
-	Accessible_unref (acc);
-    }
-    g_ptr_array_set_size (srl_watched_acc, 0);
+    if (srl_watched_acc)
+	Accessible_unref (srl_watched_acc);
+    srl_watched_acc = NULL;
 }
